//=============================================================================
// ASSEMBLE CLASS
//=============================================================================

#include "assemble.h"
#include "registers.h"

#include "streams.h"
#include "labels.h"
#include "instructions.h"
#include "directives.h"
#include "macros.h"
#include "asmclasses.h"

//-----------------------------------------------------------------------------
// ASSEMBLE::cmp: two compare instructions
//-----------------------------------------------------------------------------
BOOL ASSEMBLE::cmp( const char *a, const char *b )
{
	if(config.case_sensitive)
	{
		return((strcmp(a,b)==0)?true:false);
	} else {
		return((_stricmp(a,b)==0)?true:false);
	}
}

BOOL ASSEMBLE::cmp( const char *a, const char *b, size_t c )
{
	if(config.case_sensitive)
	{
		return((strncmp(a,b,c)==0)?true:false);
	} else {
		return((_strnicmp(a,b,c)==0)?true:false);
	}
}

//-----------------------------------------------------------------------------
// ASSEMBLE::registers: adds reg-definitions to lbl-tbl
//-----------------------------------------------------------------------------
void ASSEMBLE::init_defines()
{
    int cntr;
	if(config.default_define)	//if switch on, then include def. reg-defines
	{
        for(cntr = 32 * _R_LAST; cntr--; )
        {
            if(*r59_reg[cntr])
            {
                //if not case-sensitive, then convert 2 lower case characters
                //if(!config.case_sensitive) _strlwr(r59_reg[cntr]);
                label.add(r59_reg[cntr], cntr%32, LBL_TYPE_REG);
            }
        }
	}
}

//-----------------------------------------------------------------------------
// ASSEMBLE::detect_end: scans `til "\r\n;", set this character to zero and ret length
//-----------------------------------------------------------------------------
int ASSEMBLE::detect_end(char *str)
{
	int len = 0;
	int marks = 0;

	while(*str)
	{
		switch(*str)
        {
            case '"':
                if(marks == 1)
                {
                    if(*(str+1)=='"')
                    {
                        str++;
                    } else {
                        marks = 0;
                    }
                }
                else if(!marks) marks = 1;
                break;
            case '\'':
                if(marks == 2)
                {
                    if(*(str+1)=='\'')
                    {
                        str++;
                    } else {
                        marks=0;
                    }
                } else if(!marks) {
                    marks=2;
                }
                break;
            case ';':
            case '\r':
            case '\n':
                if(!marks)
                {
                    *str=NULL; return(len);
                }
            default:
                if(!config.case_sensitive)
                {
                    if( (!marks)&&(*str >= 'A')&&(*str <= 'Z') )
                    {
                        *str -= 'A' - 'a';
                    }
                }
        }
		str++;
		len++;
	}
	return(len);
}

//-----------------------------------------------------------------------------
// ASSEMBLE::skip_nonsens: ret ptr to first chr in str after spaces and tabs
//-----------------------------------------------------------------------------
char *ASSEMBLE::skip_nonsens(char *str)
{
    while( (*str==9)||(*str==32) )
    {
        str++;
    }

    //also delete spaces/tabs at the end
    char *end = str;
    while( *end )
    {
        end++;
    }
    while( (end>str)&&((*(end-1)==9)||(*(end-1)==32)) )
    {
        end--;
    }
    *end = 0;
    return(str);
}

//-----------------------------------------------------------------------------
// ASSEMBLE::split_str: ret ptr to following str, seperated by spaces/tabs
//-----------------------------------------------------------------------------
char *ASSEMBLE::split_str(char *str)
{
    str += strcspn(str,"\t ");			//first occurence of one of these chars
    if(!*str)
    {
        return(str);				//this is the end my friend
    }
    *str = 0;							//define end of param
    return(skip_nonsens(++str));
}

//-----------------------------------------------------------------------------
// ASSEMBLE::line: reads line for line, and splits the params
//-----------------------------------------------------------------------------
void ASSEMBLE::line(char *line)
{
    char *first,*secc,*thrd;
    DWORD ret;

    detect_end(line);
    #ifdef _DEBUG
        if(!preprocess) dbgout2("\n" << line << " -> ")
    #endif

    if(!config.case_sensitive) _strlwr(line);

    first = skip_nonsens(line);

    if(*first)	//no rem or emptyline???
    {
        if(end_of_file)
        {
            if(end_of_file > 1) return;

            end_of_file++;
            do_error(RETC_OVRWR_EOF, ERR_SHOW_POS);
            return;
		}

        secc = split_str(first);	//terminate first, and search for secc

        int label_len = strlen(first)-1;
        if(first[label_len]==':')	//if first =? label
        {
            first[label_len]=0;
            label.predef(first);
            thrd = split_str(secc);
            first = secc; secc = thrd;
        }

		if(*first)  //no rem or emptyline???
		if((ret = macro.exec(first, secc))!=RETC_OK)    //elif first =? macro
		{
			if(ret!=RETC_UNKWN_INSTR)
            {
                do_error(ret, ERR_SHOW_POS);
            } else if((ret = instruction.exec(first, secc)) != RETC_OK) {
				if(ret!=RETC_UNKWN_INSTR)
                {
                    do_error(ret, ERR_SHOW_POS);
                } else if((ret = directive.exec(NULL, first, secc))!=RETC_OK) {
					if(ret != RETC_UNKWN_INSTR)
                    {
                        do_error(ret, ERR_SHOW_POS);
                    } else {
						thrd = split_str(secc);
						if((ret = directive.exec(first, secc, thrd))!=RETC_OK)
						{
							do_error(ret, ERR_SHOW_POS);
						}
					}
				}
			}
		}
	}
	label.preset(prog_cntr, LBL_TYPE_UNDEF);
}


//-----------------------------------------------------------------------------
// ASSEMBLE::write_dbg_info: writes debug infos to file
//-----------------------------------------------------------------------------
void ASSEMBLE::write_dbg_info()
{
	if(!config.dbg_file)
    {
        return;
    }

    FILE *fh;
	DWORD cnt;
	union
	{
		FILELIST	*fl;
		CLASSLIST	*cl;
		LABELLIST	*ll;
	} ptr;

	stream.add(config.dbg_file, FTYPE_BINOUT);

    if ( !(fh = fopen(config.dbg_file, "wb")) )
    {
        do_error(ERR_FATAL|RETC_FCREATE, ERR_SHOW_FILE);
    }

    fprintf(fh, "[asm5900|dbg|v1.01]\r\n\33");
    fprintf(fh, "[file]");

    fwrite(&stream.FileList->members, sizeof(DWORD), 1, fh);
    fwrite(
        stream.FileList->h, stream.FileList->size1e,
        stream.FileList->members, fh
        );
    for(cnt = 0; cnt < stream.FileList->members; cnt++)
    {
        if(ptr.fl = (FILELIST *)stream.FileList->get(cnt))
        {
            if(ptr.fl->name)
            {
                fwrite(ptr.fl->name, 1, strlen(ptr.fl->name)+1, fh);
            } else {
                fwrite(&ptr.fl->name, 1, 1, fh);
            }
        }
    }

    fprintf(fh, "[class]");
    fwrite(&asmclass.ClassList->members, sizeof(DWORD), 1, fh);
    fwrite(
        asmclass.ClassList->h, asmclass.ClassList->size1e,
        asmclass.ClassList->members, fh
        );
    for(cnt = 0; cnt < asmclass.ClassList->members; cnt++)
    {
        if(ptr.cl = (CLASSLIST *)asmclass.ClassList->get(cnt))
        {
            if(ptr.cl->name)
            {
                fwrite(ptr.cl->name, 1, strlen(ptr.cl->name)+1, fh);
            } else {
                fwrite(&ptr.cl->name, 1, 1, fh);
            }
        }
    }

    fprintf(fh, "[label]");
    fwrite(&label.LabelList->members, sizeof(DWORD), 1, fh);
    fwrite(
        label.LabelList->h, label.LabelList->size1e,
        label.LabelList->members, fh
        );
    for(cnt = 0; cnt < label.LabelList->members; cnt++)
    {
        if(ptr.ll = (LABELLIST *)label.LabelList->get(cnt))
        {
            if(ptr.ll->name)
            {
                // debugFile.write(ptr.ll->name, strlen(ptr.ll->name)+1);
                fwrite(ptr.ll->name, 1, strlen(ptr.ll->name)+1, fh);
            } else {
                // debugFile.write(&ptr.ll->name, 1);
                fwrite(&ptr.ll->name, 1, 1, fh);
            }
        }
    }

    fprintf(fh, "[end]");
    // close(fh);
}


//-----------------------------------------------------------------------------
typedef struct
{
	char id[8];         // "PS-X EXE" or "SCE EXE"
	__uint32 fofs_text,fofs_data;	// fofs of text and data (only SCE-header)
	__uint32 entry;     // pc0 (initial program counter)
	__uint32 gp0;       // gp0 (address of the global pointer)
	__uint32 text_addr; // text_addr (address where the text sement is loaded)
	__uint32 text_size; // text_size (size of text segment)
	__uint32 data_addr,data_size;   // data_addr and data_size  (only SCE-header)
	__uint32 bss_addr,bss_size;     // bss_addr and bss_size  (only SCE-header)
	__uint32 stack_addr;            // stack address
	__uint32 stack_size;            // stack size
	//ALIGN$$	2048                // align file to segment size
} PSX_HEADER;

//-----------------------------------------------------------------------------
// ASSEMBLE::write_targ_header: writes the PSX/ELF/ECO/SCE/.. header
//-----------------------------------------------------------------------------
void ASSEMBLE::write_targ_header()
{
	//void *header;
    DWORD cnt = 0;
//    CLASSLIST *code;
    CLASSLIST *main;
    LABELLIST *lbl;
    LABELLIST *fofs;
//    LABELLIST *ofs;
//    LABELLIST *fsize;

    for(cnt = 0; cnt < label.LabelList->members; cnt++ )
    {
        lbl = (LABELLIST *)label.LabelList->get(cnt);
        // cout << "lbl " << lbl->name << " = " << lbl->value << endl;
        if ( ! strcmp(lbl->name, "_fsize") )
        {
            fofs = lbl;
            // cout << "fofs: " << fofs->value << endl;
        }
    }

    if( (lbl = (LABELLIST *)label.get("code")) )
    {
        // cout << "lbl " << lbl->name << " = " << lbl->value << endl;
    }

    for(cnt = asmclass.ClassList->members; cnt; )
    {
        if(main = (CLASSLIST *)asmclass.ClassList->get(--cnt))
        {
            if ( ! strcmp(main->name, "code") )
            {
                // cout << "code start = " << main->start << endl;
                // cout << "code fstart = " << main->fstart << endl;
                // cout << "code end = " << main->end << endl;
                // cout << "code fend = " << main->fend << endl;
                // cout << "code size = " << main->_size << endl;
                break;
            }
        }
    }

    switch(config.target)
    {
        case SEL_TARG_PSX:
            stream.out.opcode(0x582D5350, 4);
            stream.out.opcode(0x45584520, 4);
            stream.out.opcode(2048, 4);
            stream.out.opcode(0, 4);
            stream.out.opcode(0x80010000, 4);
            stream.out.opcode(0, 4);                        // gp0
            stream.out.opcode(0x80010000, 4);
            stream.out.opcode(300*1024, 4);
            stream.out.opcode(0, 4);
            stream.out.opcode(0, 4);
            stream.out.opcode(0, 4);
            stream.out.opcode(0, 4);
            stream.out.opcode(0x801FFFF0, 4);
            stream.out.opcode(0, 4);
            stream.out.align(2048);
            break;
        case SEL_TARG_ELF:
            // EHR
            stream.out.opcode(0x464C457F, 4);       // ELF Id
            stream.out.opcode(1, 1);                // File class
            stream.out.opcode(1, 1);                // Data encoding
            stream.out.opcode(1, 1);                // Header version
            stream.out.opcode(0, 8);                // nil
            stream.out.opcode(0, 1);                // nil
            stream.out.opcode(2, 2);                // ELF type
            stream.out.opcode(8, 2);                // Processor
            stream.out.opcode(1, 4);                // Version
            // Entry point addr, code.main addr
            stream.out.opcode(main->start, 4);
            stream.out.opcode(0x34, 4);             // Start of prg header
            stream.out.opcode(0x54, 4);             // Start of section header
            stream.out.opcode(0x20924001, 4);       // cpu specific flags
            stream.out.opcode(0x34, 2);             // EHR size
            stream.out.opcode(0x20, 2);             // PHR size
            stream.out.opcode(0x1, 2);              // num of PHR's
            stream.out.opcode(0x0, 2);              // SHR size
            stream.out.opcode(0x0, 2);              // num of SHR's
            stream.out.opcode(0x0, 2);
            // PHR
            stream.out.opcode(0x1, 4);              // segment type
            stream.out.opcode(0x0000007c, 4);      // fofs
            stream.out.opcode(main->start, 4);      // ofs
            stream.out.opcode(main->start, 4);      // ofs
            stream.out.opcode(0x8, 4);              // fsize
            // (code size + 15 >> 4) << 4
            stream.out.opcode(((main->_size + 15) >> 4) << 4, 4);
            stream.out.opcode(0x0, 4);              // Flags for segment
            stream.out.opcode(0x0, 4);              // Alignment
            // SHR
            stream.out.opcode(0x0, 4);              //
            stream.out.opcode(0x0, 4);
            stream.out.opcode(0x0, 4);
            stream.out.opcode(0x0, 4);
            stream.out.opcode(0x0, 4);
            stream.out.opcode(0x0, 4);              // Size of section
            stream.out.opcode(0x0, 4);
            stream.out.opcode(0x0, 4);              // Info
            stream.out.opcode(0x0, 4);              // Alignment
            stream.out.opcode(0x0, 4);
            break;
        case SEL_TARG_ECO:
            break;
        case SEL_TARG_SCE:
            break;


        case SEL_TARG_BIN:
        default:	// nothing 2 do here
            break;
    }
}

#ifdef UNIX
DWORD GetTickCount(void)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (tv.tv_sec * 1000 + tv.tv_usec / 1000);
}
#endif

//-----------------------------------------------------------------------------
// ASSEMBLE::start: does start the assembling
//-----------------------------------------------------------------------------
void ASSEMBLE::now(char *fin, char *fout)
{
    FILELIST *fl;
    char *line;
    DWORD ticks = GetTickCount();

    targ_entry = targ_gp = targ_bss = targ_stack = NULL;

    // init tables (file_no=-1 -> error detects init)
    init_defines();

    for(preprocess=true,pass=1; (pass_dependent)&&(pass < config.max_pass); pass++)
    {
        // start a new assembling process
        pass_dependent = false;
        targ_sel = false;
        end_of_file = 0;
        prog_cntr = NULL;
        config.iset = SEL_ISET_R5900;
        //targ_entry = targ_gp = targ_bss = targ_stack = NULL;
        targ_is_entry = targ_is_gp = targ_is_bss = targ_is_stack = false;

        // create outputfile
        stream.add(NULL, FTYPE_BINOUT);
        stream.out.open();

        // open inputfile
        stream.add(fin, FTYPE_DEFAULT);

        // walk through code
        while(line = stream.in.line())
        {
            dbgout("stream.in.line() -> " << line)
            assemble.line(line);
            dbgout("\n")
        }

        // close open segments
        asmclass.close_all();
        stream.out.close();
        stream.close();

        //if(error_cntr) return; //on errors stop now!
        //if(warning_cntr) return;
    }


    // do last pass (and write to file)
    // init var
    BOOL rem = pass_dependent;
    preprocess = false;
    pass_dependent = false;
    targ_sel=false;
    end_of_file = 0;
    prog_cntr = NULL;
    config.iset = SEL_ISET_R5900;
    // targ_entry = targ_gp = targ_bss = targ_stack = NULL;
    targ_is_entry = targ_is_gp = targ_is_bss = targ_is_stack = false;

    // create outputfile
    stream.add(fout, FTYPE_BINOUT);
    stream.out.open();
    if(assemble.config.lst_file)
    {
        // FILELIST *fl = stream.add(config.lst_file, FTYPE_BINOUT);
        // lst_no = stream.file_no;
        // stream.file_no = 1;
        // lst_file.open(config.lst_file, ios::write);
        if( !(lst_fh = fopen(config.lst_file, "w")) )
        {
            do_error(ERR_FATAL|RETC_FCREATE, ERR_SHOW_FILE);
        }
    }

    // init error routine
    // error_cntr = 0;
    // warning_cntr = 0;

    write_targ_header();

    // open inputfile
    stream.add(fin, FTYPE_DEFAULT);

    // walk through code
    while(line = stream.in.line())
    {
        dbgout("stream.in.line() -> " << line)
        assemble.line(line);
        dbgout("\n")
    }
    stream.file_no = 1;

    // close open segments
    asmclass.close_all();

    if(rem)
    {
        do_error(ERR_WARNING|RETC_PASS_DEPENDENT, ERR_SHOW_POS);
    }

    // close stream
    if(lst_fh)
    {
        fclose(lst_fh);
        lst_fh = NULL;
        lst_no = 0;
    }

    fl = stream.get(1);
    if( errorFile )
    {

        errorFile << fl->name << "(" << fl->line_cntr << ") - " << fl->error_cntr
            << " errors, " << fl->warning_cntr << " warnings" << endl;
    } else {
        cout << fl->name << "(" << fl->line_cntr << ") - " << fl->error_cntr
            << " errors, " << fl->warning_cntr << " warnings" << endl;
    }
    stream.out.close();
    ticks = GetTickCount() - ticks;
    fl = stream.get(0);

    if((fl)&&(fl->name)&&(*fl->name))
    {
        cout << endl << fl->name << " - ";
        if( errorFile )
        {
            errorFile << endl << fl->name << " - ";
        }
    } else {
        cout << endl;
        if( errorFile )
        {
            errorFile << endl;
        }
    }

    cout << error_cntr << " errors, " << warning_cntr << " warnings, " << pass
        << " passes in " << ticks/1000 << "." << (ticks%1000)/10 << "sec" << endl;

    if( errorFile )
    {
        errorFile << error_cntr << " errors, " << warning_cntr << " warnings, "
            << pass << " passes in " << ticks/1000 << "." << (ticks%1000)/10
            << "sec" << endl;
    }

    // write debug info to file
    write_dbg_info();

    stream.close();	//close streams
}


//-----------------------------------------------------------------------------
// ASSEMBLE::constructor
//-----------------------------------------------------------------------------
ASSEMBLE::ASSEMBLE()
{
    config.case_sensitive = true;
    config.default_define = true;
    config.default_local = true;
    config.paramtype_check = true;

    binary = true;
    config.target = SEL_TARG_DEFAULT;
    config.iset = SEL_ISET_R5900;

    config.dbg_file = NULL;
    config.lst_file = NULL;

    config.max_pass = 4;

    lst_fh = NULL;
    lst_no = (DWORD) - 1;
}

//-----------------------------------------------------------------------------
// STREAM::destructor
//-----------------------------------------------------------------------------
ASSEMBLE::~ASSEMBLE()
{
}
